@@ -31,9 +31,10 @@ For now, this module is only used on NetBSD/OpenBSD.
 
 On Linux/Android/FreeBSD, we use auxv.rs and this module is test-only because:
 - On Linux/Android, this approach requires a higher kernel version than Rust supports,
-  and also does not work with qemu-user (as of 7.2) and Valgrind (as of 3.19).
+  and also does not work with qemu-user (as of 7.2) and Valgrind (as of 3.24).
   (Looking into HWCAP_CPUID in auxvec, it appears that Valgrind is setting it
   to false correctly, but qemu-user is setting it to true.)
+  - qemu-user issue seem to be fixed as of 9.2.
 - On FreeBSD, this approach does not work on FreeBSD 12 on QEMU (confirmed on
   FreeBSD 12.{2,3,4}), and we got SIGILL (worked on FreeBSD 13 and 14).
 */
@@ -44,31 +45,60 @@ include!("common.rs");
 struct AA64Reg {
     aa64isar0: u64,
     aa64isar1: u64,
+    #[cfg(test)]
+    aa64isar3: u64,
     aa64mmfr2: u64,
 }
 
 #[cold]
 fn _detect(info: &mut CpuInfo) {
-    let AA64Reg { aa64isar0, aa64isar1, aa64mmfr2 } = imp::aa64reg();
+    let AA64Reg {
+        aa64isar0,
+        aa64isar1,
+        #[cfg(test)]
+        aa64isar3,
+        aa64mmfr2,
+    } = imp::aa64reg();
 
     // ID_AA64ISAR0_EL1, AArch64 Instruction Set Attribute Register 0
     // https://developer.arm.com/documentation/ddi0601/2024-12/AArch64-Registers/ID-AA64ISAR0-EL1--AArch64-Instruction-Set-Attribute-Register-0
+    // Atomic, bits [23:20]
+    // > FEAT_LSE implements the functionality identified by the value 0b0010.
+    // > FEAT_LSE128 implements the functionality identified by the value 0b0011.
+    // > From Armv8.1, the value 0b0000 is not permitted.
     let atomic = extract(aa64isar0, 23, 20);
     if atomic >= 0b0010 {
-        info.set(CpuInfo::HAS_LSE);
+        info.set(CpuInfoFlag::lse);
         if atomic >= 0b0011 {
-            info.set(CpuInfo::HAS_LSE128);
+            info.set(CpuInfoFlag::lse128);
         }
     }
     // ID_AA64ISAR1_EL1, AArch64 Instruction Set Attribute Register 1
     // https://developer.arm.com/documentation/ddi0601/2024-12/AArch64-Registers/ID-AA64ISAR1-EL1--AArch64-Instruction-Set-Attribute-Register-1
+    // LRCPC, bits [23:20]
+    // > FEAT_LRCPC implements the functionality identified by the value 0b0001.
+    // > FEAT_LRCPC2 implements the functionality identified by the value 0b0010.
+    // > FEAT_LRCPC3 implements the functionality identified by the value 0b0011.
+    // > From Armv8.3, the value 0b0000 is not permitted.
+    // > From Armv8.4, the value 0b0001 is not permitted.
     if extract(aa64isar1, 23, 20) >= 0b0011 {
-        info.set(CpuInfo::HAS_RCPC3);
+        info.set(CpuInfoFlag::rcpc3);
+    }
+    // ID_AA64ISAR3_EL1, AArch64 Instruction Set Attribute Register 3
+    // https://developer.arm.com/documentation/ddi0601/2024-12/AArch64-Registers/ID-AA64ISAR3-EL1--AArch64-Instruction-Set-Attribute-Register-3
+    // LSFE, bits [19:16]
+    // > FEAT_LSFE implements the functionality identified by the value 0b0001
+    #[cfg(test)]
+    if extract(aa64isar3, 19, 16) >= 0b0001 {
+        info.set(CpuInfoFlag::lsfe);
     }
     // ID_AA64MMFR2_EL1, AArch64 Memory Model Feature Register 2
     // https://developer.arm.com/documentation/ddi0601/2024-12/AArch64-Registers/ID-AA64MMFR2-EL1--AArch64-Memory-Model-Feature-Register-2
+    // AT, bits [35:32]
+    // > FEAT_LSE2 implements the functionality identified by the value 0b0001.
+    // > From Armv8.4, the value 0b0000 is not permitted.
     if extract(aa64mmfr2, 35, 32) >= 0b0001 {
-        info.set(CpuInfo::HAS_LSE2);
+        info.set(CpuInfoFlag::lse2);
     }
 }
 
@@ -92,23 +122,45 @@ mod imp {
         unsafe {
             let aa64isar0: u64;
             asm!(
-                "mrs {0}, ID_AA64ISAR0_EL1",
+                "mrs {}, ID_AA64ISAR0_EL1",
                 out(reg) aa64isar0,
                 options(pure, nomem, nostack, preserves_flags),
             );
             let aa64isar1: u64;
             asm!(
-                "mrs {0}, ID_AA64ISAR1_EL1",
+                "mrs {}, ID_AA64ISAR1_EL1",
                 out(reg) aa64isar1,
                 options(pure, nomem, nostack, preserves_flags),
             );
+            #[cfg(test)]
+            #[cfg(not(portable_atomic_pre_llvm_18))]
+            let aa64isar3: u64;
+            // ID_AA64ISAR3_EL1 is only recognized on LLVM 18+.
+            // https://github.com/llvm/llvm-project/commit/17baba9fa2728b1b1134f9dccb9318debd5a9a1b
+            #[cfg(test)]
+            #[cfg(not(portable_atomic_pre_llvm_18))]
+            asm!(
+                "mrs {}, ID_AA64ISAR3_EL1",
+                out(reg) aa64isar3,
+                options(pure, nomem, nostack, preserves_flags),
+            );
             let aa64mmfr2: u64;
             asm!(
-                "mrs {0}, ID_AA64MMFR2_EL1",
+                "mrs {}, ID_AA64MMFR2_EL1",
                 out(reg) aa64mmfr2,
                 options(pure, nomem, nostack, preserves_flags),
             );
-            AA64Reg { aa64isar0, aa64isar1, aa64mmfr2 }
+            AA64Reg {
+                aa64isar0,
+                aa64isar1,
+                #[cfg(test)]
+                #[cfg(not(portable_atomic_pre_llvm_18))]
+                aa64isar3,
+                #[cfg(test)]
+                #[cfg(portable_atomic_pre_llvm_18)]
+                aa64isar3: 0,
+                aa64mmfr2,
+            }
         }
     }
 }
@@ -200,6 +252,8 @@ mod imp {
         Some(AA64Reg {
             aa64isar0: buf.ac_aa64isar0,
             aa64isar1: buf.ac_aa64isar1,
+            #[cfg(test)]
+            aa64isar3: 0,
             aa64mmfr2: buf.ac_aa64mmfr2,
         })
     }
@@ -213,7 +267,13 @@ mod imp {
         // https://github.com/golang/sys/commit/ef9fd89ba245e184bdd308f7f2b4f3c551fa5b0f
         match sysctl_cpu_id(c!("machdep.cpu0.cpu_id")) {
             Some(cpu_id) => cpu_id,
-            None => AA64Reg { aa64isar0: 0, aa64isar1: 0, aa64mmfr2: 0 },
+            None => AA64Reg {
+                aa64isar0: 0,
+                aa64isar1: 0,
+                #[cfg(test)]
+                aa64isar3: 0,
+                aa64mmfr2: 0,
+            },
         }
     }
 }
@@ -272,7 +332,13 @@ mod imp {
         let aa64isar0 = sysctl64(&[ffi::CTL_MACHDEP, ffi::CPU_ID_AA64ISAR0]).unwrap_or(0);
         let aa64isar1 = sysctl64(&[ffi::CTL_MACHDEP, ffi::CPU_ID_AA64ISAR1]).unwrap_or(0);
         let aa64mmfr2 = sysctl64(&[ffi::CTL_MACHDEP, ffi::CPU_ID_AA64MMFR2]).unwrap_or(0);
-        AA64Reg { aa64isar0, aa64isar1, aa64mmfr2 }
+        AA64Reg {
+            aa64isar0,
+            aa64isar1,
+            #[cfg(test)]
+            aa64isar3: 0,
+            aa64mmfr2,
+        }
     }
 
     fn sysctl64(mib: &[ffi::c_int]) -> Option<u64> {
@@ -312,42 +378,22 @@ mod imp {
 )]
 #[cfg(test)]
 mod tests {
-    use std::{
-        process::Command,
-        string::{String, ToString as _},
-    };
-
     use super::*;
 
     #[test]
-    #[cfg_attr(portable_atomic_test_outline_atomics_detect_false, ignore)]
+    #[cfg_attr(portable_atomic_test_detect_false, ignore = "detection disabled")]
     fn test_aa64reg() {
-        let AA64Reg { aa64isar0, aa64isar1, aa64mmfr2 } = imp::aa64reg();
-        std::eprintln!("aa64isar0={}", aa64isar0);
-        std::eprintln!("aa64isar1={}", aa64isar1);
-        std::eprintln!("aa64mmfr2={}", aa64mmfr2);
-        if cfg!(target_os = "openbsd") {
-            let output = Command::new("sysctl").arg("machdep").output().unwrap();
-            assert!(output.status.success());
-            let stdout = String::from_utf8(output.stdout).unwrap();
-            // OpenBSD 7.1+
-            assert_eq!(
-                stdout.lines().find_map(|s| s.strip_prefix("machdep.id_aa64isar0=")).unwrap_or("0"),
-                aa64isar0.to_string(),
-            );
-            assert_eq!(
-                stdout.lines().find_map(|s| s.strip_prefix("machdep.id_aa64isar1=")).unwrap_or("0"),
-                aa64isar1.to_string(),
-            );
-            // OpenBSD 7.3+
-            assert_eq!(
-                stdout.lines().find_map(|s| s.strip_prefix("machdep.id_aa64mmfr2=")).unwrap_or("0"),
-                aa64mmfr2.to_string(),
-            );
-        }
+        let AA64Reg { aa64isar0, aa64isar1, aa64isar3, aa64mmfr2 } = imp::aa64reg();
+        test_helper::eprintln_nocapture!(
+            "aa64isar0={},aa64isar1={},aa64isar3={},aa64mmfr2={}",
+            aa64isar0,
+            aa64isar1,
+            aa64isar3,
+            aa64mmfr2,
+        );
         let atomic = extract(aa64isar0, 23, 20);
-        if detect().test(CpuInfo::HAS_LSE) {
-            if detect().test(CpuInfo::HAS_LSE128) {
+        if detect().lse() {
+            if detect().lse128() {
                 assert_eq!(atomic, 0b0011);
             } else {
                 assert_eq!(atomic, 0b0010);
@@ -356,13 +402,19 @@ mod tests {
             assert_eq!(atomic, 0b0000);
         }
         let lrcpc = extract(aa64isar1, 23, 20);
-        if detect().test(CpuInfo::HAS_RCPC3) {
+        if detect().rcpc3() {
             assert_eq!(lrcpc, 0b0011);
         } else {
             assert!(lrcpc < 0b0011, "{}", lrcpc);
         }
+        let lsfe = extract(aa64isar3, 19, 16);
+        if detect().lsfe() {
+            assert_eq!(lsfe, 0b0001);
+        } else {
+            assert_eq!(lsfe, 0b0000);
+        }
         let at = extract(aa64mmfr2, 35, 32);
-        if detect().test(CpuInfo::HAS_LSE2) {
+        if detect().lse2() {
             assert_eq!(at, 0b0001);
         } else {
             assert_eq!(at, 0b0000);
@@ -386,7 +438,7 @@ mod tests {
         //
         // This is currently used only for testing.
         fn sysctl_cpu_id_no_libc(name: &[&[u8]]) -> Result<AA64Reg, c_int> {
-            // https://github.com/golang/go/blob/4badad8d477ffd7a6b762c35bc69aed82faface7/src/syscall/asm_netbsd_arm64.s
+            // https://github.com/golang/go/blob/go1.24.0/src/syscall/asm_netbsd_arm64.s
             #[inline]
             unsafe fn sysctl(
                 name: *const c_int,
@@ -420,7 +472,7 @@ mod tests {
                 }
             }
 
-            // https://github.com/golang/sys/blob/4badad8d477ffd7a6b762c35bc69aed82faface7/cpu/cpu_netbsd_arm64.go.
+            // https://github.com/golang/sys/blob/v0.31.0/cpu/cpu_netbsd_arm64.go
             fn sysctl_nodes(mib: &mut Vec<i32>) -> Result<Vec<sys::sysctlnode>, i32> {
                 mib.push(sys::CTL_QUERY);
                 let mut q_node = sys::sysctlnode {
@@ -492,6 +544,7 @@ mod tests {
             Ok(AA64Reg {
                 aa64isar0: buf.ac_aa64isar0,
                 aa64isar1: buf.ac_aa64isar1,
+                aa64isar3: 0,
                 aa64mmfr2: buf.ac_aa64mmfr2,
             })
         }
@@ -501,4 +554,38 @@ mod tests {
             sysctl_cpu_id_no_libc(&[b"machdep", b"cpu0", b"cpu_id"]).unwrap()
         );
     }
+    #[cfg(target_os = "openbsd")]
+    #[test]
+    fn test_alternative() {
+        use std::{format, process::Command, string::String};
+
+        // Call sysctl command instead of libc API.
+        //
+        // This is used only for testing.
+        struct SysctlMachdepOutput(String);
+        impl SysctlMachdepOutput {
+            fn new() -> Self {
+                let output = Command::new("sysctl").arg("machdep").output().unwrap();
+                assert!(output.status.success());
+                let stdout = String::from_utf8(output.stdout).unwrap();
+                Self(stdout)
+            }
+            fn field(&self, name: &str) -> Option<u64> {
+                Some(
+                    self.0
+                        .lines()
+                        .find_map(|s| s.strip_prefix(&format!("{}=", name)))?
+                        .parse()
+                        .unwrap(),
+                )
+            }
+        }
+
+        let AA64Reg { aa64isar0, aa64isar1, aa64isar3, aa64mmfr2 } = imp::aa64reg();
+        let sysctl_output = SysctlMachdepOutput::new();
+        assert_eq!(aa64isar0, sysctl_output.field("machdep.id_aa64isar0").unwrap_or(0));
+        assert_eq!(aa64isar1, sysctl_output.field("machdep.id_aa64isar1").unwrap_or(0));
+        assert_eq!(aa64isar3, sysctl_output.field("machdep.id_aa64isar3").unwrap_or(0));
+        assert_eq!(aa64mmfr2, sysctl_output.field("machdep.id_aa64mmfr2").unwrap_or(0));
+    }
 }
